require_relative 'prettyprinter.rb'

# The main module that holds every SHAUN related classes and some helper functions
module Shaun
  # Helper module for classes that are tranparent SHAUN values (String, Boolean, Object)
  # It provides a default #to_sn method
  module ShaunValue
    # Cast to a SHAUN object.
    def to_sn
      self
    end
  end

  # A SHAUN object. SHAUN objects are like Ruby hashes except they only manipulate SHAUN values.
  class Object
    include ShaunValue

    # Create a SHAUN Object from a Ruby hash.
    # Every value in the hash is casted into a SHAUN value with the #to_sn method.
    # Object values are stored as instance attributes
    #
    # === Example
    # obj = Shaun::Object.new({ greetings: 'hello' })
    # greetings = obj.greetings
    # obj.greetings = 'New hello!'
    def initialize(hash = {})
      hash.each_pair do |k,v|
        name = k.to_s
        metaclass.class_eval { attr_reader name.to_sym }
        create_writer name.to_sym
        instance_variable_set "@#{name}".to_sym, v.to_sn
      end
    end

    # Add a new attribute to the object if the method name matches an attribute writer (i.e. #whatever=)
    # or calls the superclass' #method_missing method
    def method_missing(m, *params)
      res = m.match /([a-zA-Z_][a-zA-Z0-9_]*)=/
      if res and params.length == 1
        name = res[1].to_s
        metaclass.class_eval { attr_reader name.to_sym }
        create_writer name.to_sym
        instance_variable_set "@#{name}".to_sym, params[0].to_sn
      else
        super m, *params
      end
    end

    # Pretty print the SHAUN value
    def pretty_print(pp)
      pp.print_object self
    end

    private
      # Get the object's metaclass
      def metaclass
        class << self ; self ; end
      end

      # Create an attribute writer for the given symbol that castes the value into a SHAUN value
      # before assignment
      def create_writer(sym)
        metaclass.class_eval do
          define_method(sym.to_s + "=") do |value|
            instance_variable_set "@#{sym.to_s}".to_sym, value.to_sn
          end
        end
      end
  end

  # A SHAUN number. SHAUN Numbers can be integers or floats indifferently. They can also have a unit
  #
  # === Example
  # 5.2 km
  # 44 kHz
  class Number
    include ShaunValue

    # Create a new number with its value and unit
    def initialize(value, unit = nil)
      @value = value
      @unit = unit
    end

    # Cast the number to a bare Ruby Integer
    def to_i
      @value
    end

    # Cast the number to a bare Ruby Float
    def to_f
      @value
    end

    # Print the number to a String
    def to_s
      if @unit
        "#{@value.to_s} #{@unit.to_s}"
        else
          "#{@value.to_s}"
      end
    end

    # Pretty print the SHAUN value
    def pretty_print(pp)
      pp.print_number self
    end
  end

  # Helper method to create a SHAUN object
  def self.object(hash)
    Object.new hash
  end

  # Helper method to create a SHAUN list
  def self.list(ary)
    ary.to_sn
  end

  # Helper method to create a SHAUN number
  def self.number(value, unit = nil)
    Number.new value, unit
  end
end

class Hash
  def to_sn
    Shaun::Object.new self
  end
end

class Numeric
  # Cast to a SHAUN number
  def to_sn
    Shaun::Number.new self
  end

  # Patched method #method_missing in order to generate a SHAUN number
  # if the method called looks like an attribute reader
  def method_missing(m, *params)
    res = m.match /([a-zA-Z_][a-zA-Z0-9_]*)/
    if res and params.empty?
      unit = res[1].to_s
      Shaun::Number.new self, unit
    else
      super m, *params
    end
  end
end

class String
  include Shaun::ShaunValue

  # Pretty print the SHAUN value
  def pretty_print(pp)
    pp.print_string self
  end
end

class TrueClass
  include Shaun::ShaunValue

  # Pretty print the SHAUN value
  def pretty_print(pp)
    pp.print_bool self
  end
end

class FalseClass
  include Shaun::ShaunValue

  # Pretty print the SHAUN value
  def pretty_print(pp)
    pp.print_bool self
  end
end

class NilClass
  include Shaun::ShaunValue

  # Pretty print the SHAUN value
  def pretty_print(pp)
    pp.print_nil
  end
end

class Array
  # Cast to a SHAUN list. Ruby Arrays are SHAUN lists but their values
  # have to be SHAUN values as well
  def to_sn
    map { |e| e.to_sn }
  end

  # Pretty print the SHAUN value
  def pretty_print(pp)
    pp.print_list self
  end
end
